#!/usr/bin/env python
# coding: utf-8

from datetime import datetime, timedelta

from collections import defaultdict
import os
import time

import cartopy.crs as ccrs
import netCDF4
import numpy as np
import shapely

from . import map
from . import misc
from . import config as cfg


np.seterr(divide='ignore', invalid='ignore')

def iter_pixel_area(pixel, x0=cfg.startlon, y0=cfg.startlat,
        dx=cfg.dlon, dy=cfg.dlat, ie=cfg.ie, je=cfg.je):
    """\
    Iterate over indices of grid cells and yield fraction of overlap
    of pixel and cell.
    """

    if not pixel.intersects(cfg.DOMAIN):
        raise StopIteration

    dx2, dy2 = 0.5 * dx, 0.5 * dy

    # polygon bounds
    minx, miny, maxx, maxy = pixel.bounds

    # round up and down
    minx = np.floor(minx / dx) * dx
    miny = np.floor(miny / dy) * dy
    maxx = np.ceil(maxx / dx) * dx + 0.5 * dx
    maxy = np.ceil(maxy / dy) * dy + 0.5 * dy

    # start point for pixel
    i0 = int(round((minx - x0) / dx))
    j0 = int(round((miny - y0) / dy))

    # iter over grid cells
    for i,x in enumerate(np.arange(minx,maxx,dx), i0):
        for j,y in enumerate(np.arange(miny,maxy,dy), j0):

            if 0 <= i < ie and 0 <= j < je:
                grid_cell = shapely.geometry.Polygon([
                    [x-dx2, y-dy2],
                    [x-dx2, y+dy2],
                    [x+dx2, y+dy2],
                    [x+dx2, y-dy2]
                ])

                # calculate fraction of intersection and yield
                if grid_cell.intersects(pixel):
                    weight = grid_cell.intersection(pixel).area / grid_cell.area
                    yield j, i, weight




def create_pixel(lon, lat, pollon=cfg.pollon, pollat=cfg.pollat):
    """\
    Create pixel polygon in rotpole coords from corner longitude
    and latitude.
    """
    inp = ccrs.PlateCarree()
    out = ccrs.RotatedPole(
        pole_latitude=pollat,
        pole_longitude=pollon
    )
    coords = out.transform_points(inp, lon, lat)
    pixel = shapely.geometry.Polygon(coords)

    return pixel



def create_satobs(fields, lon, lat):
    """\
    Create satellite observations from orbit filename.

    fields   :: COSMO fields as list of nc variables
    lon, lat :: longitude and latitude of satellite pixel corners
    """
    shape = lon.shape[:2]
    values = np.zeros((len(fields), shape[0], shape[1]))
    weights = np.zeros(shape)

    did_load_data = False

    for n,m in np.ndindex(shape):
        pixel = create_pixel(lon[n,m], lat[n,m])
        for j,i,weight in iter_pixel_area(pixel):

            if not did_load_data:
                fields = np.array([field[0] for field in fields])
                did_load_data = True

            values[:,n,m] += fields[:,j,i] * weight
            weights[n,m] += weight

    values /= weights[np.newaxis]
    values = np.ma.array(values, mask=np.isnan(values))

    return values



def read_pixels(orbit_filename):
    with netCDF4.Dataset(orbit_filename) as nc:
        time = nc.variables['time'][...]
        lon0 = nc.variables['longitude'][...]
        lat0 = nc.variables['latitude'][...]
        lonc = nc.variables['longitude_corners'][...]
        latc = nc.variables['latitude_corners'][...]
        sza =  nc.variables['solar_zenith_angle'][...]

    return time, lon0, lat0, lonc, latc, sza


def date2orbit(date, instrument=None):
    """\
    """
    cycle_duration, cycle_length = {
                       #  days, orbits in cycle
            'Sentinel_5': (29,  412),
            'Sentinel_7': (11,  164)
    }[instrument]

    days = (date - datetime(2015,1,1)).total_seconds() / 86400.0
    return round(days / cycle_duration * cycle_length)


def save_satobs(filename, instrument, utc_time, lon_center, lat_center, lon_corners, lat_corners, sza,
        values):

    nobs, nrow, ncorner = lon_corners.shape

    if os.path.exists(filename):
        os.remove(filename)

    with netCDF4.Dataset(filename, 'w', format='NETCDF4') as nc:

        # global attributes (add input data)
        nc.setncatts({
            'DESCRIPTION':  'Synthetic Level-2 data for %s' % instrument,
            'DATAORIGIN':   'COSMO simulations, orbits from SRON orbit simulator',
            'CREATOR':      'Gerrit Kuhlmann',
            'EMAIL':        'gerrit.kuhlmann@empa.ch',
            'AFFILIATION':  'Empa Duebendorf, Switzerland',
            'DATE CREATED': time.ctime(time.time()),
            'STUDY':        'ESA project - Smartcarb',
            'time':         str(utc_time),
        })

        # create dimensions
        nobs = nc.createDimension('nobs', nobs)
        nrows = nc.createDimension('nrows', nrow)
        ncorner = nc.createDimension('ncorners', ncorner)

        # create variables
        nclat = nc.createVariable('latitude', 'f4', ('nobs', 'nrows'))
        nclon = nc.createVariable('longitude', 'f4', ('nobs', 'nrows'))
        nclat_corner = nc.createVariable('latitude_corners', 'f4', ('nobs', 'nrows', 'ncorners'))
        nclon_corner = nc.createVariable('longitude_corners', 'f4', ('nobs', 'nrows', 'ncorners'))
        ncsza = nc.createVariable('solar_zenith_angle', 'f4', ('nobs', 'nrows'))

        # Assign an attribute to the dataset
        ncsza.standard_name = "solar_zenith_angle"
        ncsza.long_name = "solar zenith angle"
        ncsza.units = "degree"
        nclat.standard_name = "latitude"
        nclat.long_name = "pixel center latitude"
        nclat.units = "degrees north"
        nclon.standard_name = "longitude"
        nclon.long_name = "pixel center longitude"
        nclon.units = "degrees east"
        nclat_corner.standard_name = "latitude"
        nclat_corner.long_name = "pixel corners latitude"
        nclat_corner.units = "degrees north"
        nclon_corner.standard_name = "longitude"
        nclon_corner.long_name = "pixel corners longitude"
        nclon_corner.units = "degrees east"

        # fill variables
        nclat[:,:] = lat_center
        nclon[:,:] = lon_center
        nclat_corner[:,:,:] = lat_corners
        nclon_corner[:,:,:] = lon_corners
        ncsza[:,:] = sza

        for key in values.keys():
            v = nc.createVariable(key, 'f4', ('nobs', 'nrows'))
            v[:,:] = values[key]

            v.standard_name = 'TODO' # TODO copy from cosmo input
            v.long_name = 'TODO'

            if key == 'CLCT':
                v.units = '1'
            elif key.startswith('XCO2'):
                v.units = 'ppm'
            else:
                v.units = 'molecules cm-2'



def main(instrument, orbit_number, names, path_orbits, path_2d_fields,
        path_level2_data, lon_eq=0.0, utc=None):

    # get cycle and orbit number
    if instrument == 'Sentinel_5':
        cycle_duration = 29.0 # days
        cycle_number, number  = divmod(orbit_number, 412)
    else:
        cycle_duration = 11.0 # days
        cycle_number, number  = divmod(orbit_number, 164)

    # get lon, lat, utc_time from orbit
    orbit_filename = os.path.join(
            path_orbits, instrument,
            '%s_o%03d_l000_pixel.nc' % (instrument, number)
    )
    time, lon0, lat0, lonc, latc, sza = read_pixels(orbit_filename)

    # shift equator passing longitude, e.g. 0,90,180 or 270
    # for four satellites
    lon0 = (lon0 + lon_eq + 180.0) % 360 - 180.0
    lonc = (lonc + lon_eq + 180.0) % 360 - 180.0

    if np.any((lonc < -180.0) | (lonc > 180.0)):
        raise ValueError

    # check if any latc, lonc is in domain
    if not (lon0.min() < 19.8 and lon0.max() > 7.5):
        print('%s orbit %d with lon %d is outside domain' % (instrument, orbit_number, lon_eq))
        return

    # use closest COSMO time or provided time
    if utc is None:

        # seconds since 2015-01-01 00:00:00"
        utc = datetime(2015,1,1) + timedelta(seconds=np.mean(time))

        # add cycles
        utc += timedelta(days=cycle_number*cycle_duration)

        # round to next hour (TODO: use linear interpolation?)
        utc = datetime(utc.year, utc.month, utc.day) + timedelta(hours=round(utc.hour + utc.minute / 60.0))

    # open 2d fields
    cosmo_filename = os.path.join(path_2d_fields, utc.strftime('cosmo_2d_%Y%m%d%H.nc'))

    # create satellite observations
    try:
        with netCDF4.Dataset(cosmo_filename) as nc:
            if names is None:
                names = sorted([
                    key for key in nc.variables.keys()
                    if key.endswith('_column') or key.startswith('XCO2') or key == 'CLCT'
                ])

            variables = [nc.variables[name] for name in names]
            satobs = create_satobs(variables, lonc, latc)

    except OSError:
        print('failed to open "%s"' % cosmo_filename)
        return

    if np.all(satobs.mask):
        print('%s orbit %d with lon %d is outside domain' % (instrument, orbit_number, lon_eq))
        return

    print('%s orbit %d with lon %d is inside domain' % (instrument, orbit_number, lon_eq))

    satobs = dict((name, satobs[i]) for i,name in enumerate(names))

    # output filename
    output_filename = os.path.join(
            path_level2_data,
            '%s_%s_o%04d_l%03d_data.nc' % (instrument, utc.strftime('%Y%m%d%H'), orbit_number, lon_eq)
    )
    save_satobs(output_filename, instrument, time, lon0, lat0,
            lonc, latc, sza, satobs)

    return utc









